from PeptideBuilder.Geometry import (
    AlaGeo,
    ArgGeo,
    AsnGeo,
    AspGeo,
    CysGeo,
    GlnGeo,
    GluGeo,
    GlyGeo,
    HisGeo,
    IleGeo,
    LeuGeo,
    LysGeo,
    MetGeo,
    PheGeo,
    ProGeo,
    SerGeo,
    ThrGeo,
    TrpGeo,
    TyrGeo,
    ValGeo,
)
from PeptideBuilder import Geometry

# test all geometries for correct parameters
def test_geometry_A():
    g = Geometry.geometry("A")
    assert isinstance(g, AlaGeo)

    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.5
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_C_O_diangle == -60.5
    assert g.N_CA_C_angle == 111.068
    assert g.N_C_CA_CB_diangle == 122.686
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "A"


def test_geometry_C():
    g = Geometry.geometry("C")
    assert isinstance(g, CysGeo)

    assert g.CA_CB_SG_angle == 113.8169
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.5
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_SG_length == 1.808
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_SG_diangle == -62.2
    assert g.N_CA_C_O_diangle == -60.0
    assert g.N_CA_C_angle == 110.8856
    assert g.N_C_CA_CB_diangle == 122.5037
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "C"


def test_geometry_D():
    g = Geometry.geometry("D")
    assert isinstance(g, AspGeo)

    assert g.CA_CB_CG_OD1_diangle == -46.7
    assert g.CA_CB_CG_OD2_diangle == 133.3
    assert g.CA_CB_CG_angle == 113.06
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.51
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_OD1_angle == 119.22
    assert g.CB_CG_OD2_angle == 118.218
    assert g.CB_CG_length == 1.52
    assert g.CG_OD1_length == 1.25
    assert g.CG_OD2_length == 1.25
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG_diangle == -66.4
    assert g.N_CA_C_O_diangle == 120.0
    assert g.N_CA_C_angle == 111.03
    assert g.N_C_CA_CB_diangle == 122.82
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "D"


def test_geometry_E():
    g = Geometry.geometry("E")
    assert isinstance(g, GluGeo)

    assert g.CA_CB_CG_CD_diangle == -179.8
    assert g.CA_CB_CG_angle == 113.82
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.511
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_CD_OE1_diangle == -6.2
    assert g.CB_CG_CD_OE2_diangle == 173.8
    assert g.CB_CG_CD_angle == 113.31
    assert g.CB_CG_length == 1.52
    assert g.CD_OE1_length == 1.25
    assert g.CD_OE2_length == 1.25
    assert g.CG_CD_OE1_angle == 119.02
    assert g.CG_CD_OE2_angle == 118.08
    assert g.CG_CD_length == 1.52
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG_diangle == -63.8
    assert g.N_CA_C_O_diangle == 120.0
    assert g.N_CA_C_angle == 111.1703
    assert g.N_C_CA_CB_diangle == 122.8702
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "E"


def test_geometry_F():
    g = Geometry.geometry("F")
    assert isinstance(g, PheGeo)

    assert g.CA_CB_CG_CD1_diangle == 93.3
    assert g.CA_CB_CG_CD2_diangle == -86.7
    assert g.CA_CB_CG_angle == 113.85
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.5316
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_CD1_CE1_diangle == 180.0
    assert g.CB_CG_CD1_angle == 120.0
    assert g.CB_CG_CD2_CE2_diangle == 180.0
    assert g.CB_CG_CD2_angle == 120.0
    assert g.CB_CG_length == 1.5
    assert g.CD1_CE1_CZ_angle == 120.0
    assert g.CD1_CE1_length == 1.39
    assert g.CD2_CE2_length == 1.39
    assert g.CE1_CZ_length == 1.39
    assert g.CG_CD1_CE1_CZ_diangle == 0.0
    assert g.CG_CD1_CE1_angle == 120.0
    assert g.CG_CD1_length == 1.39
    assert g.CG_CD2_CE2_angle == 120.0
    assert g.CG_CD2_length == 1.39
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG_diangle == -64.7
    assert g.N_CA_C_O_diangle == 120.0
    assert g.N_CA_C_angle == 110.7528
    assert g.N_C_CA_CB_diangle == 122.6054
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "F"


def test_geometry_G():
    g = Geometry.geometry("G")
    assert isinstance(g, GlyGeo)

    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.5117
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_C_O_diangle == 180.0
    assert g.N_CA_C_angle == 110.8914
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "G"


def test_geometry_H():
    g = Geometry.geometry("H")
    assert isinstance(g, HisGeo)

    assert g.CA_CB_CG_CD2_diangle == 104.3
    assert g.CA_CB_CG_ND1_diangle == -75.7
    assert g.CA_CB_CG_angle == 113.74
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.4732
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_CD2_NE2_diangle == 180.0
    assert g.CB_CG_CD2_angle == 130.61
    assert g.CB_CG_ND1_CE1_diangle == 180.0
    assert g.CB_CG_ND1_angle == 122.85
    assert g.CB_CG_length == 1.49
    assert g.CD2_NE2_length == 1.35
    assert g.CG_CD2_NE2_angle == 108.5
    assert g.CG_CD2_length == 1.35
    assert g.CG_ND1_CE1_angle == 108.5
    assert g.CG_ND1_length == 1.38
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.ND1_CE1_length == 1.32
    assert g.N_CA_CB_CG_diangle == -63.2
    assert g.N_CA_C_O_diangle == 120.0
    assert g.N_CA_C_angle == 111.0859
    assert g.N_C_CA_CB_diangle == 122.6711
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "H"


def test_geometry_I():
    g = Geometry.geometry("I")
    assert isinstance(g, IleGeo)

    assert g.CA_CB_CG1_CD1_diangle == 169.8
    assert g.CA_CB_CG1_angle == 110.7
    assert g.CA_CB_CG2_angle == 110.4
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.5403
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG1_CD1_angle == 113.97
    assert g.CB_CG1_length == 1.527
    assert g.CB_CG2_length == 1.527
    assert g.CG1_CD1_length == 1.52
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG1_diangle == 59.7
    assert g.N_CA_CB_CG2_diangle == -61.6
    assert g.N_CA_C_O_diangle == -60.0
    assert g.N_CA_C_angle == 109.7202
    assert g.N_C_CA_CB_diangle == 123.2347
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "I"


def test_geometry_K():
    g = Geometry.geometry("K")
    assert isinstance(g, LysGeo)

    assert g.CA_CB_CG_CD_diangle == -178.1
    assert g.CA_CB_CG_angle == 113.83
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.54
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_CD_CE_diangle == -179.6
    assert g.CB_CG_CD_angle == 111.79
    assert g.CB_CG_length == 1.52
    assert g.CD_CE_NZ_angle == 124.79
    assert g.CD_CE_length == 1.46
    assert g.CE_NZ_length == 1.33
    assert g.CG_CD_CE_NZ_diangle == 179.6
    assert g.CG_CD_CE_angle == 111.68
    assert g.CG_CD_length == 1.52
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG_diangle == -64.5
    assert g.N_CA_C_O_diangle == 120.0
    assert g.N_CA_C_angle == 111.08
    assert g.N_C_CA_CB_diangle == 122.76
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "K"


def test_geometry_L():
    g = Geometry.geometry("L")
    assert isinstance(g, LeuGeo)

    assert g.CA_CB_CG_CD1_diangle == 174.9
    assert g.CA_CB_CG_CD2_diangle == 66.7
    assert g.CA_CB_CG_angle == 116.1
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.4647
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_CD1_angle == 110.27
    assert g.CB_CG_CD2_angle == 110.58
    assert g.CB_CG_length == 1.53
    assert g.CG_CD1_length == 1.524
    assert g.CG_CD2_length == 1.525
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG_diangle == -60.1
    assert g.N_CA_C_O_diangle == 120.0
    assert g.N_CA_C_angle == 110.8652
    assert g.N_C_CA_CB_diangle == 122.4948
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "L"


def test_geometry_M():
    g = Geometry.geometry("M")
    assert isinstance(g, MetGeo)

    assert g.CA_CB_CG_SD_diangle == -179.6
    assert g.CA_CB_CG_angle == 113.68
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.4816
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_SD_CE_diangle == 70.1
    assert g.CB_CG_SD_angle == 112.69
    assert g.CB_CG_length == 1.52
    assert g.CG_SD_CE_angle == 100.61
    assert g.CG_SD_length == 1.81
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG_diangle == -64.4
    assert g.N_CA_C_O_diangle == 120.0
    assert g.N_CA_C_angle == 110.9416
    assert g.N_C_CA_CB_diangle == 122.6733
    assert g.SD_CE_length == 1.79
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "M"


def test_geometry_N():
    g = Geometry.geometry("N")
    assert isinstance(g, AsnGeo)

    assert g.CA_CB_CG_ND2_diangle == 121.7
    assert g.CA_CB_CG_OD1_diangle == -58.3
    assert g.CA_CB_CG_angle == 112.62
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.4826
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_ND2_angle == 116.48
    assert g.CB_CG_OD1_angle == 120.85
    assert g.CB_CG_length == 1.52
    assert g.CG_ND2_length == 1.33
    assert g.CG_OD1_length == 1.23
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG_diangle == -65.5
    assert g.N_CA_C_O_diangle == -60.0
    assert g.N_CA_C_angle == 111.5
    assert g.N_C_CA_CB_diangle == 123.2254
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "N"


def test_geometry_P():
    g = Geometry.geometry("P")
    assert isinstance(g, ProGeo)

    assert g.CA_CB_CG_CD_diangle == -34.8
    assert g.CA_CB_CG_angle == 104.21
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.2945
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_CD_angle == 105.03
    assert g.CB_CG_length == 1.49
    assert g.CG_CD_length == 1.5
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG_diangle == 29.6
    assert g.N_CA_C_O_diangle == -45.0
    assert g.N_CA_C_angle == 112.7499
    assert g.N_C_CA_CB_diangle == 115.2975
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "P"


def test_geometry_Q():
    g = Geometry.geometry("Q")
    assert isinstance(g, GlnGeo)

    assert g.CA_CB_CG_CD_diangle == -69.6
    assert g.CA_CB_CG_angle == 113.75
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.5029
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_CD_NE2_diangle == 129.5
    assert g.CB_CG_CD_OE1_diangle == -50.5
    assert g.CB_CG_CD_angle == 112.78
    assert g.CB_CG_length == 1.52
    assert g.CD_NE2_length == 1.33
    assert g.CD_OE1_length == 1.24
    assert g.CG_CD_NE2_angle == 116.5
    assert g.CG_CD_OE1_angle == 120.86
    assert g.CG_CD_length == 1.52
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG_diangle == -60.2
    assert g.N_CA_C_O_diangle == 120.0
    assert g.N_CA_C_angle == 111.0849
    assert g.N_C_CA_CB_diangle == 122.8134
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "Q"


def test_geometry_R():
    g = Geometry.geometry("R")
    assert isinstance(g, ArgGeo)

    assert g.CA_CB_CG_CD_diangle == -179.2
    assert g.CA_CB_CG_angle == 113.83
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.54
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_CD_NE_diangle == -179.3
    assert g.CB_CG_CD_angle == 111.79
    assert g.CB_CG_length == 1.52
    assert g.CD_NE_CZ_NH1_diangle == 0.0
    assert g.CD_NE_CZ_NH2_diangle == 180.0
    assert g.CD_NE_CZ_angle == 124.79
    assert g.CD_NE_length == 1.46
    assert g.CG_CD_NE_CZ_diangle == -178.7
    assert g.CG_CD_NE_angle == 111.68
    assert g.CG_CD_length == 1.52
    assert g.CZ_NH1_length == 1.33
    assert g.CZ_NH2_length == 1.33
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.NE_CZ_NH1_angle == 120.64
    assert g.NE_CZ_NH2_angle == 119.63
    assert g.NE_CZ_length == 1.33
    assert g.N_CA_CB_CG_diangle == -65.2
    assert g.N_CA_C_O_diangle == 120.0
    assert g.N_CA_C_angle == 110.98
    assert g.N_C_CA_CB_diangle == 122.76
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "R"


def test_geometry_S():
    g = Geometry.geometry("S")
    assert isinstance(g, SerGeo)

    assert g.CA_CB_OG_angle == 110.773
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.5
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_OG_length == 1.417
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_OG_diangle == -63.3
    assert g.N_CA_C_O_diangle == -60.0
    assert g.N_CA_C_angle == 111.2812
    assert g.N_C_CA_CB_diangle == 122.6618
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "S"


def test_geometry_T():
    g = Geometry.geometry("T")
    assert isinstance(g, ThrGeo)

    assert g.CA_CB_CG2_angle == 111.13
    assert g.CA_CB_OG1_angle == 109.18
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.5359
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG2_length == 1.53
    assert g.CB_OG1_length == 1.43
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG2_diangle == -60.3
    assert g.N_CA_CB_OG1_diangle == 60.0
    assert g.N_CA_C_O_diangle == 120.0
    assert g.N_CA_C_angle == 110.7014
    assert g.N_C_CA_CB_diangle == 123.0953
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "T"


def test_geometry_V():
    g = Geometry.geometry("V")
    assert isinstance(g, ValGeo)

    assert g.CA_CB_CG1_angle == 110.7
    assert g.CA_CB_CG2_angle == 110.4
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.5686
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG1_length == 1.527
    assert g.CB_CG2_length == 1.527
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG1_diangle == 177.2
    assert g.N_CA_CB_CG2_diangle == -63.3
    assert g.N_CA_C_O_diangle == -60.0
    assert g.N_CA_C_angle == 109.7698
    assert g.N_C_CA_CB_diangle == 123.2347
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "V"


def test_geometry_W():
    g = Geometry.geometry("W")
    assert isinstance(g, TrpGeo)

    assert g.CA_CB_CG_CD1_diangle == 96.3
    assert g.CA_CB_CG_CD2_diangle == -83.7
    assert g.CA_CB_CG_angle == 114.1
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.5117
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_CD1_NE1_diangle == 180.0
    assert g.CB_CG_CD1_angle == 127.07
    assert g.CB_CG_CD2_CE2_diangle == 180.0
    assert g.CB_CG_CD2_CE3_diangle == 0.0
    assert g.CB_CG_CD2_angle == 126.66
    assert g.CB_CG_length == 1.5
    assert g.CD1_NE1_length == 1.38
    assert g.CD2_CE2_CZ2_CH2_diangle == 0.0
    assert g.CD2_CE2_CZ2_angle == 120.0
    assert g.CD2_CE2_length == 1.4
    assert g.CD2_CE3_CZ3_angle == 120.0
    assert g.CD2_CE3_length == 1.4
    assert g.CE2_CZ2_CH2_angle == 120.0
    assert g.CE2_CZ2_length == 1.4
    assert g.CE3_CZ3_length == 1.4
    assert g.CG_CD1_NE1_angle == 108.5
    assert g.CG_CD1_length == 1.37
    assert g.CG_CD2_CE2_CZ2_diangle == 180.0
    assert g.CG_CD2_CE2_angle == 108.5
    assert g.CG_CD2_CE3_CZ3_diangle == 180.0
    assert g.CG_CD2_CE3_angle == 133.83
    assert g.CG_CD2_length == 1.43
    assert g.CZ2_CH2_length == 1.4
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG_diangle == -66.4
    assert g.N_CA_C_O_diangle == 120.0
    assert g.N_CA_C_angle == 110.8914
    assert g.N_C_CA_CB_diangle == 122.6112
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "W"


def test_geometry_Y():
    g = Geometry.geometry("Y")
    assert isinstance(g, TyrGeo)

    assert g.CA_CB_CG_CD1_diangle == 93.1
    assert g.CA_CB_CG_CD2_diangle == 273.1
    assert g.CA_CB_CG_angle == 113.8
    assert g.CA_CB_length == 1.52
    assert g.CA_C_N_angle == 116.642992978143
    assert g.CA_C_O_angle == 120.5434
    assert g.CA_C_length == 1.52
    assert g.CA_N_length == 1.46
    assert g.CB_CG_CD1_CE1_diangle == 180.0
    assert g.CB_CG_CD1_angle == 120.98
    assert g.CB_CG_CD2_CE2_diangle == 180.0
    assert g.CB_CG_CD2_angle == 120.82
    assert g.CB_CG_length == 1.51
    assert g.CD1_CE1_CZ_OH_diangle == 180.0
    assert g.CD1_CE1_CZ_angle == 120.0
    assert g.CD1_CE1_length == 1.39
    assert g.CD2_CE2_length == 1.39
    assert g.CE1_CZ_OH_angle == 119.78
    assert g.CE1_CZ_length == 1.39
    assert g.CG_CD1_CE1_CZ_diangle == 0.0
    assert g.CG_CD1_CE1_angle == 120.0
    assert g.CG_CD1_length == 1.39
    assert g.CG_CD2_CE2_angle == 120.0
    assert g.CG_CD2_length == 1.39
    assert g.CZ_OH_length == 1.39
    assert g.C_CA_CB_angle == 109.5
    assert g.C_N_CA_angle == 121.382215820277
    assert g.C_O_length == 1.23
    assert g.N_CA_CB_CG_diangle == -64.3
    assert g.N_CA_C_O_diangle == 120.0
    assert g.N_CA_C_angle == 110.9288
    assert g.N_C_CA_CB_diangle == 122.6023
    assert g.omega == 180.0
    assert g.peptide_bond == 1.33
    assert g.phi == -120
    assert g.psi_im1 == 140
    assert g.residue_name == "Y"



""" HyResBuilder package is based on the PeptideBuilder, which was modified for HyRes model.
``PeptideBuilder`` package for creating peptide models in PDB format based on geometrical parameters
Written by Matthew Z. Tien, Dariya K. Sydykova, Austin G. Meyer, and Claus O. Wilke.
Python modules
----------------
The package consists of the following Python modules:
* HyresBuilder
* Geometry
"""
__version__ = "1.2.7"
from .HyresBuilder import *
from .Geometry import *



"""This module is part of the PeptideBuilder library,
written by Matthew Z. Tien, Dariya K. Sydykova,
Austin G. Meyer, and Claus O. Wilke, which was modified by Shanlong Li,
and named as HyresBuilder.

The Geometry module contains the default geometries of
all 20 amino acids for HyRes model. The main function to be used is the
geometry() function, which returns the default geometry
for the requested amino acid.

This file is provided to you under the MIT License."""

import random
from typing import List


class Geo:
    """Geometry base class"""

    residue_name: str

    # Geometry to bring together residue
    peptide_bond: float
    CA_C_N_angle: float
    C_N_CA_angle: float

    # Backbone coordinates
    N_CA_C_angle: float
    CA_N_length: float
    CA_C_length: float
    phi: float
    psi_im1: float
    omega: float

    # Carbonyl atom
    C_O_length: float
    CA_C_O_angle: float
    N_CA_C_O_diangle: float

    # backbone Hydrogen atom
    N_H_length: float
    CA_N_H_angle: float
    C_CA_N_H_diangle: float

    def __repr__(self) -> str:
        repr = ""
        for var in self.__dict__:
            repr += "%s = %s\n" % (var, self.__dict__[var])
        return repr


class GlyGeo(Geo):
    """Geometry of Glycine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.residue_name = "G"


class AlaGeo(Geo):
    """Geometry of Alanin"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 1.63
        self.C_CA_CB_angle = 111.2
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "A"


class SerGeo(Geo):
    """Geometry of Serine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 2.03
        self.C_CA_CB_angle = 109.4
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "S"


class CysGeo(Geo):
    """Geometry of Cystine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 2.40
        self.C_CA_CB_angle = 118.3
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "C"


class ValGeo(Geo):
    """Geometry of Valine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 2.01
        self.C_CA_CB_angle = 116.3
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "V"


class IleGeo(Geo):
    """Geometry of Isoleucine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 2.33
        self.C_CA_CB_angle = 117.8
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "I"


# The following function is commented out, because it is not
# recommended to randomize rotamers for isoleucine. The underlying
# reason for this recommendation is that isoleucine's beta-carbon
# is a chiral center.
##    def generateRandomRotamers(self):
##        rotamer_bins = [-60, 60, 180]
##        tempList = []
##        for i in range(0, 3):
##            tempList.append(random.choice(rotamer_bins))
##        self.inputRotamers(tempList)


class LeuGeo(Geo):
    """Geometry of Leucine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 2.71
        self.C_CA_CB_angle = 120.6
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "L"


class ThrGeo(Geo):
    """Geometry of Threonine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 2.01
        self.C_CA_CB_angle = 116.2
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "T"


class ArgGeo(Geo):
    """Geometry of Arginine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 2.56
        self.C_CA_CB_angle = 118.5
        self.N_C_CA_CB_diangle = 120.0

        self.CB_CC_length = 3.51
        self.CA_CB_CC_angle = 137.9
        self.N_CA_CB_CC_diangle = 0.0

        self.residue_name = "R"

    def inputRotamers(self, rotamers: List[float]) -> None:
        try:
            self.N_CA_CB_CC_diangle = rotamers[0]
        except IndexError:
            print("Input Rotamers List: not long enough")
            self.N_CA_CB_CC_diangle = 0.0


class LysGeo(Geo):
    """Geometry of Lysine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 2.57
        self.C_CA_CB_angle = 117.0
        self.N_C_CA_CB_diangle = 120.0

        self.CB_CC_length = 3.02
        self.CA_CB_CC_angle = 144.9
        self.N_CA_CB_CC_diangle = 75.0

        self.residue_name = "K"

    def inputRotamers(self, rotamers: List[float]) -> None:
        try:
            self.N_CA_CB_CC_diangle = rotamers[0]
        except IndexError:
            print("Input Rotamers List: not long enough")
            self.N_CA_CB_CC_diangle = 75.0



class AspGeo(Geo):
    """Geometry of Aspartic Acid"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 2.54
        self.C_CA_CB_angle = 118.5
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "D"


class AsnGeo(Geo):
    """Geometry of Asparagine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 2.56
        self.C_CA_CB_angle = 127.0
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "N"


class GluGeo(Geo):
    """Geometry of Glutamic Acid"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 3.14
        self.C_CA_CB_angle = 116.2
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "E"


class GlnGeo(Geo):
    """Geometry of Glutamine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 3.23
        self.C_CA_CB_angle = 119.1
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "Q"


class MetGeo(Geo):
    """Geometry of Methionine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 3.26
        self.C_CA_CB_angle = 119.8
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "M"


class HisGeo(Geo):
    """Geometry of Histidine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 1.94
        self.C_CA_CB_angle = 112.5
        self.N_C_CA_CB_diangle = 120.0

        self.CB_CC_length = 2.52
        self.CA_CB_CC_angle = 129.5
        self.N_CA_CB_CC_diangle = 0.0
        
        self.CC_CD_length = 1.86
        self.CB_CC_CD_angle = 66.0
        self.CA_CB_CC_CD_diangle = 180.0

        self.residue_name = "H"


class ProGeo(Geo):
    """Geometry of Proline"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 1.94
        self.C_CA_CB_angle = 121.7
        self.N_C_CA_CB_diangle = 120.0

        self.residue_name = "P"


class PheGeo(Geo):
    """Geometry of Phenylalanine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 2.35
        self.C_CA_CB_angle = 123.2
        self.N_C_CA_CB_diangle = 120.0

        self.CB_CC_length = 2.45
        self.CA_CB_CC_angle = 116.0
        self.N_CA_CB_CC_diangle = 0.00

        self.CC_CD_length = 2.22
        self.CB_CC_CD_angle = 75.66
        self.CA_CB_CC_CD_diangle = 180.00

        self.residue_name = "F"


class TyrGeo(Geo):
    """Geometry of Tyrosine"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 2.35
        self.C_CA_CB_angle = 123.6
        self.N_C_CA_CB_diangle = 120.0

        self.CB_CC_length = 2.45
        self.CA_CB_CC_angle = 115.8
        self.N_CA_CB_CC_diangle = 0.00

        self.CC_CD_length = 2.44
        self.CB_CC_CD_angle = 92.1
        self.CA_CB_CC_CD_diangle = 180.00

        self.residue_name = "Y"


class TrpGeo(Geo):
    """Geometry of Tryptophan"""

    def __init__(self):
        self.CA_N_length = 1.45
        self.CA_C_length = 1.52
        self.N_CA_C_angle = 111.6

        self.C_O_length = 1.23
        self.CA_C_O_angle = 121.6
        self.N_CA_C_O_diangle = 180.0
        
        self.N_H_length = 0.98
        self.CA_N_H_angle = 120.0
        self.C_CA_N_H_diangle = 180.0

        self.phi = -120
        self.psi_im1 = 140
        self.omega = 180.0
        self.peptide_bond = 1.33
        self.CA_C_N_angle = 116.642992978143
        self.C_N_CA_angle = 121.382215820277

        self.CA_CB_length = 1.99
        self.C_CA_CB_angle = 116.0
        self.N_C_CA_CB_diangle = 120.0

        self.CB_CC_length = 2.58
        self.CA_CB_CC_angle = 125.6
        self.N_CA_CB_CC_diangle = 0.00

        self.CB_CD_length = 2.53
        self.CA_CB_CD_angle = 133.7
        self.N_CA_CB_CD_diangle = 180.00
        
        self.CD_CE_length = 2.15
        self.CB_CD_CE_angle = 98.6
        self.CA_CB_CD_CE_diangle = 180.00
        
        self.CD_CF_length = 2.15
        self.CB_CD_CF_angle = 168.1
        self.CA_CB_CD_CF_diangle = 180.00

        self.residue_name = "W"


def geometry(AA: str) -> Geo:
    """Generates the geometry of the requested amino acid.
    The amino acid needs to be specified by its single-letter
    code. If an invalid code is specified, the function
    returns the geometry of Glycine."""
    if AA == "G":
        return GlyGeo()
    elif AA == "A":
        return AlaGeo()
    elif AA == "S":
        return SerGeo()
    elif AA == "C":
        return CysGeo()
    elif AA == "V":
        return ValGeo()
    elif AA == "I":
        return IleGeo()
    elif AA == "L":
        return LeuGeo()
    elif AA == "T":
        return ThrGeo()
    elif AA == "R":
        return ArgGeo()
    elif AA == "K":
        return LysGeo()
    elif AA == "D":
        return AspGeo()
    elif AA == "E":
        return GluGeo()
    elif AA == "N":
        return AsnGeo()
    elif AA == "Q":
        return GlnGeo()
    elif AA == "M":
        return MetGeo()
    elif AA == "H":
        return HisGeo()
    elif AA == "P":
        return ProGeo()
    elif AA == "F":
        return PheGeo()
    elif AA == "Y":
        return TyrGeo()
    elif AA == "W":
        return TrpGeo()
    else:
        return GlyGeo()



from setuptools import setup


INSTALL_REQUIRES = ["Biopython"]

TEST_REQUIRES = [
    # testing and coverage
    "pytest",
    "coverage",
    "pytest-cov",
    # to be able to run `python setup.py checkdocs`
    "collective.checkdocs",
    "pygments",
]


with open("README.md", "r") as f:
    long_description = f.read()

with open("HyresBuilder/__init__.py", "r") as f:
    init = f.readlines()

for line in init:
    if "__version__" in line:
        __version__ = line.split('"')[-2]

setup(
    name="HyresBuilder",
    version=__version__,
    author="Shanlong Li",
    author_email="shanlongli@umass.edu",
    description="Create HyRes peptide PDB files with specified geometry",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wayuer19/HyresBuilder",
    download_url="https://github.com/wayuer19/HyresBuilder/releases",
    platforms="Tested on Ubuntu 22.04",
    packages=["HyresBuilder"],
    install_requires=INSTALL_REQUIRES,
    extras_require={"test": TEST_REQUIRES + INSTALL_REQUIRES,},
    classifiers=[
        # Trove classifiers
        # (https://pypi.python.org/pypi?%3Aaction=list_classifiers)
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Intended Audience :: Science/Research",
    ],
)



"""This module is part of the PeptideBuilder library,
written by Matthew Z. Tien, Dariya K. Sydykova,
Austin G. Meyer, and Claus O. Wilke.

The PeptideBuilder module contains code to generate 3D
structures of peptides. It requires the Geometry module
(also part of the PeptideBuilder library), which contains
default bond lengths and angles for all amino acids.

This module also requires the Bio.PDB module from
Biopython, for structure manipulation.

This file is provided to you under the MIT License."""

import math, warnings
from typing import List, Optional, Union

from Bio.PDB.Polypeptide import is_aa
from Bio.PDB.Atom import Atom
from Bio.PDB.Residue import Residue
from Bio.PDB.Chain import Chain
from Bio.PDB.Model import Model
from Bio.PDB.Structure import Structure
from Bio.PDB.vectors import Vector, rotaxis, calc_dihedral, calc_angle
import numpy as np

from .Geometry import (
    AlaGeo,
    ArgGeo,
    AsnGeo,
    AspGeo,
    CysGeo,
    GlnGeo,
    GluGeo,
    GlyGeo,
    HisGeo,
    IleGeo,
    LeuGeo,
    LysGeo,
    MetGeo,
    PheGeo,
    ProGeo,
    SerGeo,
    ThrGeo,
    TrpGeo,
    TyrGeo,
    ValGeo,
    geometry,
    Geo,
)


def calculateCoordinates(
    refA: Residue, refB: Residue, refC: Residue, L: float, ang: float, di: float
) -> np.ndarray:
    AV = refA.get_vector()
    BV = refB.get_vector()
    CV = refC.get_vector()

    CA = AV - CV
    CB = BV - CV

    ##CA vector
    AX = CA[0]
    AY = CA[1]
    AZ = CA[2]

    ##CB vector
    BX = CB[0]
    BY = CB[1]
    BZ = CB[2]

    ##Plane Parameters
    A = (AY * BZ) - (AZ * BY)
    B = (AZ * BX) - (AX * BZ)
    G = (AX * BY) - (AY * BX)

    ##Dot Product Constant
    F = math.sqrt(BX * BX + BY * BY + BZ * BZ) * L * math.cos(ang * (math.pi / 180.0))

    ##Constants
    const = math.sqrt(
        math.pow((B * BZ - BY * G), 2)
        * (
            -(F * F) * (A * A + B * B + G * G)
            + (
                B * B * (BX * BX + BZ * BZ)
                + A * A * (BY * BY + BZ * BZ)
                - (2 * A * BX * BZ * G)
                + (BX * BX + BY * BY) * G * G
                - (2 * B * BY) * (A * BX + BZ * G)
            )
            * L
            * L
        )
    )
    denom = (
        (B * B) * (BX * BX + BZ * BZ)
        + (A * A) * (BY * BY + BZ * BZ)
        - (2 * A * BX * BZ * G)
        + (BX * BX + BY * BY) * (G * G)
        - (2 * B * BY) * (A * BX + BZ * G)
    )

    X = (
        (B * B * BX * F) - (A * B * BY * F) + (F * G) * (-A * BZ + BX * G) + const
    ) / denom

    if (B == 0 or BZ == 0) and (BY == 0 or G == 0):
        const1 = math.sqrt(
            G * G * (-A * A * X * X + (B * B + G * G) * (L - X) * (L + X))
        )
        Y = ((-A * B * X) + const1) / (B * B + G * G)
        Z = -(A * G * G * X + B * const1) / (G * (B * B + G * G))
    else:
        Y = (
            (A * A * BY * F) * (B * BZ - BY * G)
            + G * (-F * math.pow(B * BZ - BY * G, 2) + BX * const)
            - A * (B * B * BX * BZ * F - B * BX * BY * F * G + BZ * const)
        ) / ((B * BZ - BY * G) * denom)
        Z = (
            (A * A * BZ * F) * (B * BZ - BY * G)
            + (B * F) * math.pow(B * BZ - BY * G, 2)
            + (A * BX * F * G) * (-B * BZ + BY * G)
            - B * BX * const
            + A * BY * const
        ) / ((B * BZ - BY * G) * denom)

    # Get the new Vector from the origin
    D = Vector(X, Y, Z) + CV
    with warnings.catch_warnings():
        # ignore inconsequential warning
        warnings.simplefilter("ignore")
        temp = calc_dihedral(AV, BV, CV, D) * (180.0 / math.pi)

    di = di - temp
    rot = rotaxis(math.pi * (di / 180.0), CV - BV)
    D = (D - BV).left_multiply(rot) + BV

    return D.get_array()


def makeGly(segID: int, N, H, CA, C, O, geo: Geo) -> Residue:
    """Creates a Glycine residue"""
    res = Residue((" ", segID, " "), "GLY", "    ")

    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(C)
    res.add(O)
    return res


def makeAla(segID: int, N, H, CA, C, O, geo: AlaGeo) -> Residue:
    """Creates an Alanine residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")

    res = Residue((" ", segID, " "), "ALA", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makeSer(segID: int, N, H, CA, C, O, geo: SerGeo) -> Residue:
    """Creates a Serine residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")
    
    ##Create Reside Data Structure
    res = Residue((" ", segID, " "), "SER", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makeCys(segID: int, N, H, CA, C, O, geo: CysGeo) -> Residue:
    """Creates a Cysteine residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle
    
    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")

    res = Residue((" ", segID, " "), "CYS", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makeVal(segID: int, N, H, CA, C, O, geo: ValGeo) -> Residue:
    """Creates a Valine residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")

    res = Residue((" ", segID, " "), "VAL", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makeIle(segID: int, N, H, CA, C, O, geo: IleGeo) -> Residue:
    """Creates an Isoleucine residue"""
    ##R-group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle
    
    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")

    res = Residue((" ", segID, " "), "ILE", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makeLeu(segID: int, N, H, CA, C, O, geo: LeuGeo) -> Residue:
    """Creates a Leucine residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")

    res = Residue((" ", segID, " "), "LEU", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makeThr(segID: int, N, H, CA, C, O, geo: ThrGeo) -> Residue:
    """Creates a Threonine residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")

    res = Residue((" ", segID, " "), "THR", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makeArg(segID: int, N, H, CA, C, O, geo: ArgGeo) -> Residue:
    """Creates an Arginie residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    CB_CC_length = geo.CB_CC_length
    CA_CB_CC_angle = geo.CA_CB_CC_angle
    N_CA_CB_CC_diangle = geo.N_CA_CB_CC_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")
    carbon_g = calculateCoordinates(
        N, CA, CB, CB_CC_length, CA_CB_CC_angle, N_CA_CB_CC_diangle
    )
    CC = Atom("CC", carbon_g, 0.0, 1.0, " ", " CC", 0, "C")

    res = Residue((" ", segID, " "), "ARG", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(CC)
    res.add(C)
    res.add(O)
    return res


def makeLys(segID: int, N, H, CA, C, O, geo: LysGeo) -> Residue:
    """Creates a Lysine residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    CB_CC_length = geo.CB_CC_length
    CA_CB_CC_angle = geo.CA_CB_CC_angle
    N_CA_CB_CC_diangle = geo.N_CA_CB_CC_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")
    carbon_g = calculateCoordinates(
        N, CA, CB, CB_CC_length, CA_CB_CC_angle, N_CA_CB_CC_diangle
    )
    CC = Atom("CC", carbon_g, 0.0, 1.0, " ", " CC", 0, "C")

    res = Residue((" ", segID, " "), "LYS", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(CC)
    res.add(C)
    res.add(O)
    return res


def makeAsp(segID: int, N, H, CA, C, O, geo: AspGeo) -> Residue:
    """Creates an Aspartic Acid residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")

    res = Residue((" ", segID, " "), "ASP", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makeAsn(segID, N, H, CA, C, O, geo):
    """Creates an Asparagine residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")
    
    res = Residue((" ", segID, " "), "ASN", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makeGlu(segID: int, N, H, CA, C, O, geo: GluGeo) -> Residue:
    """Creates a Glutamic Acid residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")

    res = Residue((" ", segID, " "), "GLU", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makeGln(segID: int, N, H, CA, C, O, geo: GlnGeo) -> Residue:
    """Creates a Glutamine residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")

    res = Residue((" ", segID, " "), "GLN", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makeMet(segID: int, N, H, CA, C, O, geo: MetGeo) -> Residue:
    """Creates a Methionine residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")

    res = Residue((" ", segID, " "), "MET", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makeHis(segID: int, N, H, CA, C, O, geo: HisGeo) -> Residue:
    """Creates a Histidine residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    CB_CC_length = geo.CB_CC_length
    CA_CB_CC_angle = geo.CA_CB_CC_angle
    N_CA_CB_CC_diangle = geo.N_CA_CB_CC_diangle

    CC_CD_length = geo.CC_CD_length
    CB_CC_CD_angle = geo.CB_CC_CD_angle
    CA_CB_CC_CD_diangle = geo.CA_CB_CC_CD_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")
    carbon_g = calculateCoordinates(
        N, CA, CB, CB_CC_length, CA_CB_CC_angle, N_CA_CB_CC_diangle
    )
    CC = Atom("CC", carbon_g, 0.0, 1.0, " ", " CC", 0, "C")
    carbon_d2 = calculateCoordinates(
        CA, CB, CC, CC_CD_length, CB_CC_CD_angle, CA_CB_CC_CD_diangle
    )
    CD = Atom("CD", carbon_d2, 0.0, 1.0, " ", " CD", 0, "C")

    res = Residue((" ", segID, " "), "HIS", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(CC)
    res.add(CD)
    res.add(C)
    res.add(O)
    return res


def makePro(segID: int, N, H, CA, C, O, geo: ProGeo) -> Residue:
    """Creates a Proline residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")

    res = Residue((" ", segID, " "), "PRO", "    ")
    res.add(N)
    res.add(CA)
    res.add(CB)
    res.add(C)
    res.add(O)
    return res


def makePhe(segID: int, N, H, CA, C, O, geo: PheGeo) -> Residue:
    """Creates a Phenylalanine residue"""
    ##R-Group    
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    CB_CC_length = geo.CB_CC_length
    CA_CB_CC_angle = geo.CA_CB_CC_angle
    N_CA_CB_CC_diangle = geo.N_CA_CB_CC_diangle

    CC_CD_length = geo.CC_CD_length
    CB_CC_CD_angle = geo.CB_CC_CD_angle
    CA_CB_CC_CD_diangle = geo.CA_CB_CC_CD_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")
    carbon_g = calculateCoordinates(
        N, CA, CB, CB_CC_length, CA_CB_CC_angle, N_CA_CB_CC_diangle
    )
    CC = Atom("CC", carbon_g, 0.0, 1.0, " ", " CC", 0, "C")
    carbon_d2 = calculateCoordinates(
        CA, CB, CC, CC_CD_length, CB_CC_CD_angle, CA_CB_CC_CD_diangle
    )
    CD = Atom("CD", carbon_d2, 0.0, 1.0, " ", " CD", 0, "C")

    res = Residue((" ", segID, " "), "PHE", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(CC)
    res.add(CD)
    res.add(C)
    res.add(O)
    return res


def makeTyr(segID: int, N, H, CA, C, O, geo: TyrGeo) -> Residue:
    """Creates a Tyrosine residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    CB_CC_length = geo.CB_CC_length
    CA_CB_CC_angle = geo.CA_CB_CC_angle
    N_CA_CB_CC_diangle = geo.N_CA_CB_CC_diangle

    CC_CD_length = geo.CC_CD_length
    CB_CC_CD_angle = geo.CB_CC_CD_angle
    CA_CB_CC_CD_diangle = geo.CA_CB_CC_CD_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")
    carbon_g = calculateCoordinates(
        N, CA, CB, CB_CC_length, CA_CB_CC_angle, N_CA_CB_CC_diangle
    )
    CC = Atom("CC", carbon_g, 0.0, 1.0, " ", " CC", 0, "C")
    carbon_d2 = calculateCoordinates(
        CA, CB, CC, CC_CD_length, CB_CC_CD_angle, CA_CB_CC_CD_diangle
    )
    CD = Atom("CD", carbon_d2, 0.0, 1.0, " ", " CD", 0, "C")

    res = Residue((" ", segID, " "), "TYR", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(CC)
    res.add(CD)
    res.add(C)
    res.add(O)
    return res


def makeTrp(segID: int, N, H, CA, C, O, geo: TrpGeo) -> Residue:
    """Creates a Tryptophan residue"""
    ##R-Group
    CA_CB_length = geo.CA_CB_length
    C_CA_CB_angle = geo.C_CA_CB_angle
    N_C_CA_CB_diangle = geo.N_C_CA_CB_diangle

    CB_CC_length = geo.CB_CC_length
    CA_CB_CC_angle = geo.CA_CB_CC_angle
    N_CA_CB_CC_diangle = geo.N_CA_CB_CC_diangle

    CB_CD_length = geo.CB_CD_length
    CA_CB_CD_angle = geo.CA_CB_CD_angle
    N_CA_CB_CD_diangle = geo.N_CA_CB_CD_diangle

    CD_CE_length = geo.CD_CE_length
    CB_CD_CE_angle = geo.CB_CD_CE_angle
    CA_CB_CD_CE_diangle = geo.CA_CB_CD_CE_diangle

    CD_CF_length = geo.CD_CF_length
    CB_CD_CF_angle = geo.CB_CD_CF_angle
    CA_CB_CD_CF_diangle = geo.CA_CB_CD_CF_diangle

    carbon_b = calculateCoordinates(
        N, C, CA, CA_CB_length, C_CA_CB_angle, N_C_CA_CB_diangle
    )
    CB = Atom("CB", carbon_b, 0.0, 1.0, " ", " CB", 0, "C")
    carbon_g = calculateCoordinates(
        N, CA, CB, CB_CC_length, CA_CB_CC_angle, N_CA_CB_CC_diangle
    )
    CC = Atom("CC", carbon_g, 0.0, 1.0, " ", " CC", 0, "C")
    carbon_d1 = calculateCoordinates(
        N, CA, CB, CB_CD_length, CA_CB_CD_angle, N_CA_CB_CD_diangle
    )
    CD = Atom("CD", carbon_d1, 0.0, 1.0, " ", " CD", 0, "C")
    carbon_d2 = calculateCoordinates(
        CA, CB, CD, CD_CE_length, CB_CD_CE_angle, CA_CB_CD_CE_diangle
    )
    CE = Atom("CE", carbon_d2, 0.0, 1.0, " ", " CE", 0, "C")
    carbon_e2 = calculateCoordinates(
        CA, CB, CD, CD_CF_length, CB_CD_CF_angle, CA_CB_CD_CF_diangle
    )
    CF = Atom("CF", carbon_e2, 0.0, 1.0, " ", " CF", 0, "C")

    ##Create Residue DS
    res = Residue((" ", segID, " "), "TRP", "    ")
    res.add(N)
    res.add(H)
    res.add(CA)
    res.add(CB)
    res.add(CC)
    res.add(CD)
    res.add(CE)
    res.add(CF)
    res.add(C)
    res.add(O)
    return res


def make_res_of_type(segID: int, N, H, CA, C, O, geo: Geo) -> Residue:
    if isinstance(geo, GlyGeo):
        res = makeGly(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, AlaGeo):
        res = makeAla(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, SerGeo):
        res = makeSer(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, CysGeo):
        res = makeCys(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, ValGeo):
        res = makeVal(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, IleGeo):
        res = makeIle(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, LeuGeo):
        res = makeLeu(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, ThrGeo):
        res = makeThr(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, ArgGeo):
        res = makeArg(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, LysGeo):
        res = makeLys(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, AspGeo):
        res = makeAsp(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, GluGeo):
        res = makeGlu(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, AsnGeo):
        res = makeAsn(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, GlnGeo):
        res = makeGln(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, MetGeo):
        res = makeMet(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, HisGeo):
        res = makeHis(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, ProGeo):
        res = makePro(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, PheGeo):
        res = makePhe(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, TyrGeo):
        res = makeTyr(segID, N, H, CA, C, O, geo)
    elif isinstance(geo, TrpGeo):
        res = makeTrp(segID, N, H, CA, C, O, geo)
    else:
        res = makeGly(segID, N, H, CA, C, O, geo)

    return res


def initialize_res(residue: Union[Geo, str]) -> Structure:
    """Creates a new structure containing a single amino acid. The type and
    geometry of the amino acid are determined by the argument, which has to be
    either a geometry object or a single-letter amino acid code.
    The amino acid will be placed into chain A of model 0."""

    if isinstance(residue, Geo):
        geo = residue
    elif isinstance(residue, str):
        geo = geometry(residue)
    else:
        raise ValueError("Invalid residue argument:", residue)

    segID = 1
    AA = geo.residue_name
    CA_N_length = geo.CA_N_length
    CA_C_length = geo.CA_C_length
    N_CA_C_angle = geo.N_CA_C_angle

    CA_coord = np.array([0.0, 0.0, 0.0])
    C_coord = np.array([CA_C_length, 0, 0])
    N_coord = np.array(
        [
            CA_N_length * math.cos(N_CA_C_angle * (math.pi / 180.0)),
            CA_N_length * math.sin(N_CA_C_angle * (math.pi / 180.0)),
            0,
        ]
    )

    N = Atom("N", N_coord, 0.0, 1.0, " ", " N", 0, "N")
    CA = Atom("CA", CA_coord, 0.0, 1.0, " ", " CA", 0, "C")
    C = Atom("C", C_coord, 0.0, 1.0, " ", " C", 0, "C")

    ## Create Hydrogen atom
    N_H_length = geo.N_H_length
    CA_N_H_angle = geo.CA_N_H_angle
    C_CA_N_H_diangle = geo.C_CA_N_H_diangle
    hydrogen = calculateCoordinates(C, CA, N, N_H_length, CA_N_H_angle, C_CA_N_H_diangle)
    H = Atom("H", hydrogen, 0.0, 1.0, " ", " H", 0, "H")
    
    ##Create Carbonyl atom (to be moved later)
    C_O_length = geo.C_O_length
    CA_C_O_angle = geo.CA_C_O_angle
    N_CA_C_O_diangle = geo.N_CA_C_O_diangle

    carbonyl = calculateCoordinates(
        N, CA, C, C_O_length, CA_C_O_angle, N_CA_C_O_diangle
    )
    O = Atom("O", carbonyl, 0.0, 1.0, " ", " O", 0, "O")

    res = make_res_of_type(segID, N, H, CA, C, O, geo)

    cha = Chain("A")
    cha.add(res)

    mod = Model(0)
    mod.add(cha)

    struc = Structure("X")
    struc.add(mod)
    return struc


def getReferenceResidue(structure: Structure) -> Residue:
    """Returns the last residue of chain A model 0 of the given structure.

    This function is a helper function that should not normally be called
    directly."""

    # If the following line doesn't work we're in trouble.
    # Likely initialize_res() wasn't called.
    resRef = structure[0]["A"].child_list[-1]

    # If the residue is not an amino acid we're in trouble.
    # Likely somebody is trying to append residues to an existing
    # structure that has non-amino-acid molecules in the chain.
    assert is_aa(resRef)

    return resRef


def add_residue_from_geo(structure: Structure, geo: Geo) -> Structure:
    """Adds a residue to chain A model 0 of the given structure, and
    returns the new structure. The residue to be added is determined by
    the geometry object given as second argument.

    This function is a helper function and should not normally be called
    directly. Call add_residue() instead."""
    resRef = getReferenceResidue(structure)
    AA = geo.residue_name
    segID = resRef.get_id()[1]
    segID += 1

    ##geometry to bring together residue
    peptide_bond = geo.peptide_bond
    CA_C_N_angle = geo.CA_C_N_angle
    C_N_CA_angle = geo.C_N_CA_angle

    ##Backbone Coordinates
    N_CA_C_angle = geo.N_CA_C_angle
    CA_N_length = geo.CA_N_length
    CA_C_length = geo.CA_C_length
    phi = geo.phi
    psi_im1 = geo.psi_im1
    omega = geo.omega

    N_coord = calculateCoordinates(resRef["N"], resRef["CA"], resRef["C"], peptide_bond, CA_C_N_angle, psi_im1)
    N = Atom("N", N_coord, 0.0, 1.0, " ", " N", 0, "N")

    CA_coord = calculateCoordinates(resRef["CA"], resRef["C"], N, CA_N_length, C_N_CA_angle, omega)
    CA = Atom("CA", CA_coord, 0.0, 1.0, " ", " CA", 0, "C")

    C_coord = calculateCoordinates(resRef["C"], N, CA, CA_C_length, N_CA_C_angle, phi)
    C = Atom("C", C_coord, 0.0, 1.0, " ", " C", 0, "C")
    
    ##Create Carbonyl atom (to be moved later)
    C_O_length = geo.C_O_length
    CA_C_O_angle = geo.CA_C_O_angle
    N_CA_C_O_diangle = geo.N_CA_C_O_diangle

    carbonyl = calculateCoordinates(N, CA, C, C_O_length, CA_C_O_angle, N_CA_C_O_diangle)
    O = Atom("O", carbonyl, 0.0, 1.0, " ", " O", 0, "O")

    if AA != 'P':
        ## Create Hydrogen atom
        N_H_length = geo.N_H_length
        CA_N_H_angle = geo.CA_N_H_angle
        C_CA_N_H_diangle = geo.C_CA_N_H_diangle
    
        hydrogen = calculateCoordinates(C, CA, N, N_H_length, CA_N_H_angle, C_CA_N_H_diangle)
        H = Atom("H", hydrogen, 0.0, 1.0, " ", " H", 0, "H")
    else:
        H = O
    
    res = make_res_of_type(segID, N, H, CA, C, O, geo)

    resRef["O"].set_coord(
        calculateCoordinates(
            res["N"], resRef["CA"], resRef["C"], C_O_length, CA_C_O_angle, 180.0
        )
    )

    ghost = Atom(
        "N",
        calculateCoordinates(
            res["N"], res["CA"], res["C"], peptide_bond, CA_C_N_angle, psi_im1
        ),
        0.0,
        0.0,
        " ",
        "N",
        0,
        "N",
    )
    res["O"].set_coord(
        calculateCoordinates(
            res["N"], res["CA"], res["C"], C_O_length, CA_C_O_angle, 180.0
        )
    )

    structure[0]["A"].add(res)
    return structure


def make_extended_structure(AA_chain: str) -> Structure:
    """Place a sequence of amino acids into a peptide in the extended
    conformation. The argument AA_chain holds the sequence of amino
    acids to be used."""
    geo = geometry(AA_chain[0])
    struc = initialize_res(geo)

    for i in range(1, len(AA_chain)):
        AA = AA_chain[i]
        geo = geometry(AA)
        add_residue(struc, geo)

    return struc


def add_residue(
    structure: Structure, residue: Union[Geo, str], phi=-120, psi_im1=140, omega=-370
) -> Structure:
    """Adds a residue to chain A model 0 of the given structure, and
    returns the new structure. The residue to be added can be specified
    in two ways: either as a geometry object (in which case
    the remaining arguments phi, psi_im1, and omega are ignored) or as a
    single-letter amino-acid code. In the latter case, the optional
    arguments phi, psi_im1, and omega specify the corresponding backbone
    angles.

    When omega is specified, it needs to be a value greater than or equal
    to -360. Values below -360 are ignored."""

    if isinstance(residue, Geo):
        geo = residue
    elif isinstance(residue, str):
        geo = geometry(residue)
        geo.phi = phi
        geo.psi_im1 = psi_im1
        if omega > -361:
            geo.omega = omega
    else:
        raise ValueError("Invalid residue argument:", residue)

    return add_residue_from_geo(structure, geo)


def make_structure(
    AA_chain: str, phi: List[float], psi_im1: List[float], omega: Optional[List] = None
) -> Structure:
    """Place a sequence of amino acids into a peptide with specified
    backbone dihedral angles. The argument AA_chain holds the
    sequence of amino acids to be used. The arguments phi and psi_im1 hold
    lists of backbone angles, one for each amino acid, *starting from
    the second amino acid in the chain*. The argument
    omega (optional) holds a list of omega angles, also starting from
    the second amino acid in the chain."""
    geo = geometry(AA_chain[0])
    struc = initialize_res(geo)

    if omega is None or not len(omega):
        for i in range(1, len(AA_chain)):
            AA = AA_chain[i]
            add_residue(struc, AA, phi[i - 1], psi_im1[i - 1])
    else:
        for i in range(1, len(AA_chain)):
            AA = AA_chain[i]
            add_residue(struc, AA, phi[i - 1], psi_im1[i - 1], omega[i - 1])

    return struc


def make_structure_from_geos(geos: List[Geo]) -> Structure:
    """Creates a structure out of a list of geometry objects."""
    model_structure = initialize_res(geos[0])
    for i in range(1, len(geos)):
        add_residue(model_structure, geos[i])

    return model_structure


def add_terminal_OXT(structure: Structure, C_OXT_length: float = 1.23) -> Structure:
    """Adds a terminal oxygen atom ('OXT') to the last residue of chain A model 0 of the given structure, and returns the new structure. The OXT atom object will be contained in the last residue object of the structure.

This function should be used only when the structure object is completed and no further residues need to be appended."""

    rad = 180.0 / math.pi

    # obtain last residue infomation
    resRef = getReferenceResidue(structure)
    N_resRef = resRef["N"]
    CA_resRef = resRef["CA"]
    C_resRef = resRef["C"]
    O_resRef = resRef["O"]

    n_vec = N_resRef.get_vector()
    ca_vec = CA_resRef.get_vector()
    c_vec = C_resRef.get_vector()
    o_vec = O_resRef.get_vector()

    # geometry to bring together residue
    CA_C_OXT_angle = calc_angle(ca_vec, c_vec, o_vec) * rad
    N_CA_C_O_diangle = calc_dihedral(n_vec, ca_vec, c_vec, o_vec) * rad
    N_CA_C_OXT_diangle = N_CA_C_O_diangle - 180.0
    if N_CA_C_O_diangle < 0:
        N_CA_C_OXT_diangle = N_CA_C_O_diangle + 180.0

    # OXT atom creation
    OXT_coord = calculateCoordinates(
        N_resRef, CA_resRef, C_resRef, C_OXT_length, CA_C_OXT_angle, N_CA_C_OXT_diangle
    )
    OXT = Atom("OXT", OXT_coord, 0.0, 1.0, " ", "OXT", 0, "O")

    # modify last residue of the structure to contain the OXT atom
    resRef.add(OXT)
    return structure
